@@ -41,7 +41,17 @@ struct Inner<FE> {
     size: usize,
 }
 
-struct Cache<FE> {
+impl<FE> Inner<FE> {
+    fn new() -> Self {
+        Self {
+            size: 0,
+            files: HashMap::new(),
+        }
+    }
+}
+
+/// An in-memory cache layer over [`tokio::fs`] with least-frequently-used (LFU) eviction.
+pub struct Cache<FE> {
     capacity: usize,
     lfu: LFU,
     inner: Mutex<Inner<FE>>,
@@ -58,15 +68,6 @@ impl<FE> Cache<FE> {
         }
     }
 
-    fn remove(&self, path: &PathBuf, entry_size: usize) {
-        let mut state = self.inner.lock().expect("file cache state");
-
-        if state.files.remove(path).is_some() {
-            self.lfu.remove(path);
-            state.size -= entry_size;
-        }
-    }
-
     fn resize(&self, old_size: usize, new_size: usize) {
         let mut state = self.inner.lock().expect("file cache state");
 
@@ -78,56 +79,66 @@ impl<FE> Cache<FE> {
     }
 }
 
-impl<FE> Cache<FE> {
-    fn new(capacity: usize) -> Self {
-        let inner = Mutex::new(Inner {
-            size: 0,
-            files: HashMap::new(),
+impl<FE: FileLoad + Send + Sync + 'static> Cache<FE> {
+    /// Initialize the cache.
+    ///
+    /// `cleanup_interval` specifies how often cache cleanup should run in the background.
+    ///
+    /// This function should only be called once.
+    pub fn new(capacity: usize, cleanup_interval: Duration) -> Arc<Self> {
+        let cache = Arc::new(Self {
+            lfu: LFU::new(),
+            inner: Mutex::new(Inner::new()),
+            capacity,
         });
 
-        Self {
-            lfu: LFU::new(),
-            inner,
-            capacity,
-        }
-    }
-}
+        spawn_cleanup_thread(cache.clone(), cleanup_interval);
 
-/// Load the filesystem cache from the given `root` directory.
-///
-/// `duration` specified how frequently the background cleanup thread should check if the
-/// cache is full.
-pub async fn load<FE: FileLoad + Send + Sync + 'static>(
-    root: PathBuf,
-    cache_size: usize,
-    cleanup_interval: Duration,
-) -> Result<DirLock<FE>, io::Error> {
-    let cache = Arc::new(Cache::new(cache_size));
-    spawn_cleanup_thread(cache.clone(), cleanup_interval);
-    let dir = DirLock::load(cache, root).await?;
-    Ok(dir)
+        cache
+    }
+
+    /// Load a filesystem directory into the cache.
+    ///
+    /// After loading, all interactions with files under this directory should go through
+    /// a [`DirLock`] or [`FileLock`].
+    pub async fn load(self: Arc<Self>, path: PathBuf) -> Result<DirLock<FE>, io::Error> {
+        {
+            let lock = self.inner.lock().expect("file cache state");
+            for file_path in lock.files.keys() {
+                if file_path.starts_with(&path) {
+                    return Err(io::Error::new(
+                        io::ErrorKind::AlreadyExists,
+                        format!(
+                            "called Cache::load on a directory that's already loaded: {:?}",
+                            path
+                        ),
+                    ));
+                }
+            }
+        }
+
+        DirLock::load(self, path).await
+    }
 }
 
 fn spawn_cleanup_thread<FE: FileLoad + Send + Sync + 'static>(
     cache: Arc<Cache<FE>>,
     interval: Duration,
 ) -> tokio::task::JoinHandle<()> {
-    let mut interval = tokio::time::interval(interval);
-
     tokio::spawn(async move {
         loop {
             loop {
                 if cache.inner.lock().expect("file cache state").size > cache.capacity {
                     break;
                 } else {
-                    interval.tick().await;
+                    tokio::time::sleep(interval).await;
                 }
             }
 
             let mut evictions = {
+                let state = cache.inner.lock().expect("file cache state");
+                let mut over = state.size as i64 - cache.capacity as i64;
                 let evictions = FuturesUnordered::new();
-                let occupied = cache.inner.lock().expect("file cache state").size;
-                let mut over = occupied as i64 - cache.capacity as i64;
 
                 let mut lfu = cache.lfu.iter();
                 while let Some(path) = lfu.next() {
@@ -135,18 +146,11 @@ fn spawn_cleanup_thread<FE: FileLoad + Send + Sync + 'static>(
                         break;
                     }
 
-                    let state = cache.inner.lock().expect("file cache state");
                     if let Some(file) = state.files.get(&path).cloned() {
-                        if let Some(size) = file.size() {
-                            if let Some(eviction) = file.evict() {
-                                evictions.push(eviction);
-                                over -= size as i64;
-                            }
+                        if let Some((size, eviction)) = file.evict() {
+                            over -= size as i64;
+                            evictions.push(eviction);
                         }
-                    } else {
-                        // since we're not holding the lock on `state`,
-                        // the file may already have been removed
-                        // between calling lfu.iter() and now
                     }
                 }
 
@@ -155,12 +159,12 @@ fn spawn_cleanup_thread<FE: FileLoad + Send + Sync + 'static>(
 
             while let Some(result) = evictions.next().await {
                 match result {
-                    Ok(()) => {}
+                    Ok(()) => {},
                     Err(cause) => panic!("failed to evict file from cache: {}", cause),
                 }
             }
 
-            interval.tick().await;
+            tokio::time::sleep(interval).await
         }
     })
 }
